/*
 *	coretrace - basic debugger for embedded Linux
 *	Copyright (C) 2004 Ronny L Nilsson
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
 *
 */



/***********************************************************
 * Generic CORE file handler
 *
 **********************************************************/

#include <sys/types.h>
#include <asm/string.h>
#include <sys/procfs.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <stdarg.h>

#include "core-linux.h"
#include "main.h"
#include "exe-elf.h"



//====== DEFINITIONS =================


//====== GLOBALS ==================
const char *stackName="stack";									// Description of the stack memmap			
struct core_t *core;											// Descriptor of the corefile		
long pageSize;													// Systems PAGE_SIZE		



//====== PROTOTYPES ================
int coreElf(void);
int corePhrd(void);
int coreNotes(void);
int core2ExeMap(void);

extern int coreGetArchSummary(void);							// defined by every arch
extern int coreArchStackLocate(void);							// defined by every arch




//----------------------------------------------------------
// Init core analyzer
//----------------------------------------------------------
int coreInit(void) {
	int errLoc;

	errLoc=0;
	pageSize=0;
	core=NULL;
	
	// Get architecture page size
	if(!errLoc) {
		pageSize=sysconf(_SC_PAGESIZE);
		if(pageSize==-1) {
			perror("Error getting page size");
			errLoc=1;
			pageSize=0;
		}
	}

	return errLoc;
}




//----------------------------------------------------------
// Open corefile and map into memory.
//----------------------------------------------------------
int coreOpen(const char *corePath) {
	int errLoc, res;
	struct stat coreStat;
	char resolvedPath[PATH_MAX];
	
	errLoc=0;
	memset(&coreStat, 0, sizeof(coreStat));
	memset(resolvedPath, 0, sizeof(resolvedPath));
	
	if(corePath && strlen(corePath)) {
		if(realpath(corePath, resolvedPath) == NULL) {							// Get absolute name	
			perror("Error accessing corefile");
			errLoc=1;
		}
	}
	else {
		errLoc=1;
	}
			
	if(!errLoc && strlen(resolvedPath) && 
			access(resolvedPath, F_OK | R_OK)==0) {
		core=malloc(sizeof(struct core_t));
		if(core && errno != ENOMEM) {
			memset(core, 0, sizeof(struct core_t));
			strncpy(core->path, resolvedPath, PATH_MAX);
		}
		else {
			fprintf(stderr, "Error getting memory for core\n");
			errLoc=1;
		}
		
		// Get file size
		if(!errLoc) {
			res=stat(core->path, &coreStat);
			if(res==0) {
				if(S_ISREG(coreStat.st_mode)) {
					core->fileSize=coreStat.st_size;
				}
				else {
					fprintf(stderr, "Error, wrong corefile type\n");			// Error if not a regular file	
					errLoc=1;
				}
			}
			else if(res==-1) {
				perror("Error accessing corefile");
				errLoc=1;
			}
		}
		
		// Map file into mem
		if(!errLoc) {
			core->fd=open(core->path, O_RDONLY);
			if(core->fd==-1) {
				perror("Error opening corefile");
				errLoc=1;
			}
		}		
		if(!errLoc && core->fd) {
			if(core->fileSize % pageSize) {
				core->dataLength = core->fileSize + (pageSize - 				// Roundup to page size 
					(core->fileSize % pageSize));
			}
			else {
				core->dataLength = core->fileSize;
			}
			core->data=mmap(0, core->dataLength, PROT_READ, 
				MAP_SHARED, core->fd, 0);
			if(core->data==MAP_FAILED) {
				fprintf(stderr, "Error mapping corefile\n");
				errLoc=1;
				core->data=NULL;
			}
		}
	}
	else {
		core=NULL;
	}
	
	// Restore on errors
	if(errLoc) coreClose();
	

	return errLoc;
}







//----------------------------------------------------------
// Cleanup core analyser, return 0 on success.
//----------------------------------------------------------
int coreClose(void) {	
	struct map_t *map, *mapPrev;
	int errLoc;
	
	errLoc=0;
	
	if(core) {
		if(core->maps) {
			map=core->maps;
			while(map->next) {
				mapPrev=map;			
				map=map->next;
				free(mapPrev->name);
				free(mapPrev);
			}
			free(map->name);
			free(map);
			core->maps=NULL;
			core->stack=NULL;
		}
			
		if(core->data) {
			munmap(core->data, core->dataLength); 
			core->data=NULL;
			core->dataLength=0;
		}

		if(core->fd) {
			close(core->fd);
			core->fd=0;
		}

		free(core);
		core=NULL;
	}

	return errLoc;
}




//----------------------------------------------------------
// Return a human readable string, telling breifely
// what's in the corefile and what was the malfunction
// creating it. Note, caller must inspect string length!
// I may be zero. Also, caller must copy it to preserve it,
// it'll be erased in crashCoreClose().
//----------------------------------------------------------
char* coreGetSummary(void) {
	int errLoc, res;
	
	errLoc=0;
	
	if(core) {
		memset(core->summary, 0, MAX_SUMMARY_LENGTH);
		core->summaryLength=0;
	
		if(!errLoc) errLoc=coreElf();
		if(!errLoc) errLoc=corePhrd();
		if(!errLoc) errLoc=coreNotes();
		if(!errLoc) errLoc=core2ExeMap();
		if(!errLoc) errLoc=coreArchStackLocate();
		
		// Show general, common info
		if(!errLoc && core->psinfo) {
			res=snprintf(core->summary + core->summaryLength, 
				MAX_SUMMARY_LENGTH - core->summaryLength, 
				"Application: %s\n", core->psinfo->pr_psargs);
			if(res>0) core->summaryLength+=res;
		}
		if(!errLoc && core->prstatus && core->prstatus->pr_cursig) {
			res=snprintf(core->summary + core->summaryLength, 
				MAX_SUMMARY_LENGTH - core->summaryLength, 
				"Offending signal: %hd ", core->prstatus->pr_cursig);
			if(res>0) core->summaryLength+=res;
			res=0;
			switch(core->prstatus->pr_cursig) {
				case SIGSEGV:
					res = snprintf(core->summary + core->summaryLength, 
						MAX_SUMMARY_LENGTH - core->summaryLength, 
						"(SIGSEGV)\n");
					break;

				case SIGABRT:
					res = snprintf(core->summary + core->summaryLength, 
						MAX_SUMMARY_LENGTH - core->summaryLength, 
						"(SIGABRT)\n");
					break;

				case SIGFPE:
					res = snprintf(core->summary + core->summaryLength, 
						MAX_SUMMARY_LENGTH - core->summaryLength, 
						"(SIGFPE)\n");
					break;

				case SIGQUIT:
					res = snprintf(core->summary + core->summaryLength, 
						MAX_SUMMARY_LENGTH - core->summaryLength, 
						"(SIGQUIT)\n");
					break;

				default:
					res = snprintf(core->summary + core->summaryLength, 
						MAX_SUMMARY_LENGTH - core->summaryLength, 
						"(please look it up bits/signum.h)\n");
					break;
			}
			if(res>0) core->summaryLength+=res;
		}
		
		// Give warning if hasn't found a stack map
		if(!errLoc && !core->stack) {
			res=snprintf(core->summary + core->summaryLength, 
				MAX_SUMMARY_LENGTH - core->summaryLength, 
				"Warning, no stack found!\n");
			if(res>0) core->summaryLength+=res;
		}		

		// Finaly, show architecture dependant info
		if(!errLoc) errLoc=coreGetArchSummary();
		
		// Show a tooltip
		if(!errLoc && core->summaryLength) {
			res=snprintf(core->summary + core->summaryLength, 
				MAX_SUMMARY_LENGTH - core->summaryLength, 
				"Convert these numbers with 'addr2line'.\n");
			if(res>0) core->summaryLength+=res;
		}		

		return core->summary;
	}
	else {
		return NULL;
	}
}



//----------------------------------------------------------
// Parse standard elf header from a corefile
// buffer <coreData>.
// Returns 0 on success.
//----------------------------------------------------------
int coreElf(void) {
	int errLoc;
	
	core->elf = core->data;
	errLoc=0;
	
	/* Read ELF header */
	if(!core->data || core->dataLength<sizeof(Elf32_Ehdr) || 
			memcmp(core->data, ELFMAG, SELFMAG)!=0 || 
			core->elf->e_type!=ET_CORE) {
		fprintf(stderr, "Wrong elf core format!\n");
		errLoc=1;
	}

	if(!errLoc) {
		switch(core->elf->e_machine) {
#if defined(EM_CRIS) && defined(__CRIS__)
			case EM_CRIS:
				DBG("Parsing cris corefile\n");
				break;
#elif defined(EM_386) && defined(__i386__)
			case EM_386:
				DBG("Parsing x86 corefile\n");
				break;
#elif defined(EM_PPC) && defined(__PPC__)
			case EM_PPC:
				if(core->elf->e_ident[EI_DATA]==ELFDATA2MSB) {
					DBG("Parsing ppc corefile\n");
				}
				else {
					DBG("PPC32 little endian corefile (yet unsupported!)\n");
					errLoc=1;
				}
				break;
#else
#error No supported arch configured!
#endif
			default:
				fprintf(stderr, "Error, unknown architecture!");
				errLoc=1;
				break;
		}
	}
	
	return errLoc;
}
	

	
//----------------------------------------------------------
// Parse standard elf program headers section
// Returns 0 on success.
//----------------------------------------------------------
int corePhrd(void) {
	int errLoc, prgHdrIdx;
	struct map_t *map;
	
	prgHdrIdx=0;
	errLoc=0;
	map=NULL;
	
	/* Read program headers */
	if(core->elf && core->elf->e_phnum) {
		core->prgHdr = (Elf32_Phdr*) (core->data + core->elf->e_phoff);
		DBG("Program headers of %s:\n", core->path);
		for(prgHdrIdx=0; prgHdrIdx<core->elf->e_phnum && !errLoc; prgHdrIdx++) {
			switch((core->prgHdr+prgHdrIdx)->p_type) {
				case PT_NOTE:
					if(core->notes) {
						DBG("\tNotes at   0x%8.8x of 0x%8.8x bytes. "
							" Warning, ignored since we've already got one!\n",
							(core->prgHdr+prgHdrIdx)->p_offset,
							(core->prgHdr+prgHdrIdx)->p_filesz);
					}
					else {
						core->notes = (Elf32_Nhdr*) (core->data + 
							(core->prgHdr+prgHdrIdx)->p_offset);
						core->notesLength=(core->prgHdr+prgHdrIdx)->p_filesz;
						DBG("\tNotes at   0x%8.8x of 0x%8.8x bytes\n",
							(core->prgHdr+prgHdrIdx)->p_offset,
							(core->prgHdr+prgHdrIdx)->p_filesz);
					}
					break;

				case PT_LOAD:
					DBG("\tMem map at 0x%8.8x of 0x%8.8x bytes\n",
						(core->prgHdr+prgHdrIdx)->p_offset,
						(core->prgHdr+prgHdrIdx)->p_filesz);
						
					if(core->maps) {
						map=core->maps;					
						while(map->next) map=map->next;
						map->next=calloc(1, sizeof(struct map_t));
						if(!map->next) {
							fprintf(stderr, "Error getting mem for map\n");
							errLoc=1;
						}
						map=map->next;
					}
					else {
						core->maps=calloc(1, sizeof(struct map_t));
						if(!core->maps) {
							fprintf(stderr, "Error getting mem for map\n");
							errLoc=1;
						}
						map=core->maps;
					}
					if(!errLoc && map) map->hdr=core->prgHdr+prgHdrIdx;
					break;

				default:
					break;
			}
		}
	}

	return errLoc;
}



//----------------------------------------------------------
// Parse standard elf program headers notes section
// Returns 0 on success.
//----------------------------------------------------------
int coreNotes(void) {
	Elf32_Nhdr *note;
	Elf32_Word offs;
	int errLoc, nameIdx;
	
	errLoc=0;	
	offs=0;
	
	/* Read the notes */
	if(core->notes && core->notesLength) {
		DBG("Notes of %s:\n", core->path);
		while(offs<core->notesLength) {
			note = (Elf32_Nhdr*) (((void*) core->notes) + offs);
			offs += sizeof(Elf32_Nhdr);

			DBG("\tName: ");
			for(nameIdx=0; nameIdx<note->n_namesz; nameIdx++) {					// Print the note names
				DBG("%c", *((char*) (((void*) core->notes) + 
					offs + nameIdx)));
			}
			offs +=note->n_namesz;

			DBG("\tContains: ");
			switch(note->n_type) {
				case NT_PRSTATUS:
					DBG("process status");
					core->prstatus=(struct elf_prstatus*) 
						(((void*) core->notes) + offs);
					core->regs=(user_regs_struct_t*) 
						&core->prstatus->pr_reg;
					break;

				case NT_PRPSINFO:
					DBG("process state");
					core->psinfo=(struct elf_prpsinfo*)
						(((void*) core->notes) + offs);
					break;
				
				case NT_TASKSTRUCT:
					DBG("task struct");
					break;

				default:
					break;
			}			
			DBG("\t");
			DBG("\n");
			offs += roundup(note->n_descsz, 4);									// Add up size of note data length		
		}
	}

	return errLoc;
}





//----------------------------------------------------------
// Map names from binary executable mappings to the
// corresponding core mappings.
//----------------------------------------------------------
int core2ExeMap(void) {
	int errLoc;
	struct map_t *mapExe, *mapCore;
	errLoc=0;
	

	if(exe && core) {
		mapExe=exe->maps;
		while(!errLoc && mapExe) {
			mapCore=core->maps;
			while(!errLoc && mapCore) {
				if(mapExe->hdr->p_vaddr==mapCore->hdr->p_vaddr) {
					memcpy(mapExe->hdr, mapCore->hdr, sizeof(Elf32_Phdr));
					mapCore->isDynamic=mapExe->isDynamic;
					if(strlen(mapExe->name)) {
						mapCore->name=malloc(strlen(mapExe->name)+1);
						if(mapCore->name) strcpy(mapCore->name, mapExe->name);
					}
				}
				mapCore=mapCore->next;
			}
			mapExe=mapExe->next;
		}
	}
	
	return errLoc;
}


