/*
 *	coretrace - basic debugger for embedded Linux
 *	Copyright (C) 2004 Ronny L Nilsson
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
 *
 */


 
/***********************************************************
 * Generic ELF executable file handler
 *
 **********************************************************/
 
#include <sys/types.h>
#include <asm/string.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/wait.h>

#include "exe-elf.h"
#include "main.h"



//====== DEFINITIONS =================



//====== GLOBALS ==================
struct exe_t *exe;												// Descriptor of the binary who crached		

//====== PROTOTYPES ================
int exeElf(void);
int exePhrd(void);
int exeNameDynamics(void);



//----------------------------------------------------------
// Open binary executable
//----------------------------------------------------------
int exeOpen(const char *exePath) {
	int errLoc, res;
	struct stat exeStat;
	
	errLoc=0;
	memset(&exeStat, 0, sizeof(exeStat));
		
	if(exePath && strlen(exePath) && access(exePath, R_OK | X_OK)==0) {
		exe=malloc(sizeof(struct exe_t));
		if(exe && strlen(exePath)<PATH_MAX) {
			memset(exe, 0, sizeof(struct exe_t));
						
			if(realpath(exePath, exe->path)==NULL) {							// Get absolute name	
				perror("Error resolving exe path");
				errLoc=1;
			}			
		}
		else {
			fprintf(stderr, "Error getting mem for exe\n");
			errLoc=1;
		}

		// Get file size
		if(!errLoc) {
			res=stat(exe->path, &exeStat);
			if(res==0 && S_ISREG(exeStat.st_mode)) {
				exe->fileSize=exeStat.st_size;
			}
			else if(res==-1) {
				perror("Error accessing executable");
				errLoc=1;
			}
		}
		
		// Map file into mem
		if(!errLoc) {
			exe->fd=open(exe->path, O_RDONLY);
			if(exe->fd==-1) {
				perror("Error opening executable");
				errLoc=1;
			}
		}		
		if(!errLoc && exe->fd) {
			if(exe->fileSize % pageSize) {
				exe->dataLength = exe->fileSize + (pageSize - 					// Roundup to page size 
					(exe->fileSize % pageSize));
			}
			else {
				exe->dataLength = exe->fileSize;
			}
			exe->data=mmap(0, exe->dataLength, PROT_READ, 
				MAP_SHARED, exe->fd, 0);
			if(exe->data==MAP_FAILED) {
				fprintf(stderr, "Error mapping executable\n");
				errLoc=1;
				exe->data=NULL;
			}
		}
	}
	else {
		exe=NULL;
	}
	
	// Restore on errors
	if(errLoc) exeClose();
	
	return errLoc;
	
}



//----------------------------------------------------------
// Close and cleanup exe parser
//----------------------------------------------------------
int exeClose(void) {
	int errLoc;
	struct map_t *map, *mapPrev;
	
	errLoc=0;

	if(exe) {
		if(exe->maps) {
			map=exe->maps;
			while(map->next) {
				mapPrev=map;			
				map=map->next;				
				free(mapPrev->hdr);
				free(mapPrev->name);
				free(mapPrev);
			}
			free(map->hdr);
			free(map->name);
			free(map);
			exe->maps=NULL;
		}
		
		if(exe->data) {
			munmap(exe->data, exe->dataLength);
			exe->data=NULL;
			exe->dataLength=0;
		}
	
		if(exe->fd) {
			close(exe->fd);
			exe->fd=0;
		}
		
		free(exe);
		exe=NULL;
	}
	
	return errLoc;
}




//----------------------------------------------------------
// Parse and get summary out of binary executable
//----------------------------------------------------------
int exeGetSummary(void) {
	int errLoc;
	
	errLoc=0;
	
	if(exe && exe->data) {
		if(!errLoc) errLoc=exeElf();
		if(!errLoc) errLoc=exePhrd();
		if(!errLoc) errLoc=exeNameDynamics();
	}
	
	return errLoc;
}




//----------------------------------------------------------
// Parse standard elf header from executable
// Returns 0 on success.
//----------------------------------------------------------
int exeElf(void) {
	int errLoc;
	
	exe->elf = exe->data;
	errLoc=0;
	
	/* Read ELF header */
	if(!exe->data || exe->dataLength<sizeof(Elf32_Ehdr) || 
			memcmp(exe->data, ELFMAG, SELFMAG)!=0 || 
			exe->elf->e_type!=ET_EXEC) {
		fprintf(stderr, "Wrong executable format!\n");
		errLoc=1;
	}
	
	return errLoc;
}
	


//----------------------------------------------------------
// Parse standard elf program headers section
// Returns 0 on success.
//----------------------------------------------------------
int exePhrd(void) {
	int errLoc, prgHdrIdx;
	struct map_t *map;
	
	errLoc=0;
	
	/* Read program headers */
	if(exe->elf && exe->elf->e_phnum) {
		exe->prgHdr = (Elf32_Phdr*) (exe->data + exe->elf->e_phoff);
		DBG("Program headers of %s:\n", exe->path);
		for(prgHdrIdx=0; prgHdrIdx<exe->elf->e_phnum && !errLoc; prgHdrIdx++) {
			switch((exe->prgHdr+prgHdrIdx)->p_type) {
				case PT_INTERP:
					DBG("\tLinking by %s\n", 
						(char*)(exe->data + 
						(exe->prgHdr+prgHdrIdx)->p_offset));
					exe->isDynamic=1;
					break;

				case PT_NOTE:
					DBG("\tNotes at   0x%8.8x of 0x%8.8x bytes\n",
						(exe->prgHdr+prgHdrIdx)->p_offset,
						(exe->prgHdr+prgHdrIdx)->p_filesz);
					break;
					
				case PT_LOAD:
					DBG("\tCode at    0x%8.8x of 0x%8.8x bytes\n",
						(exe->prgHdr+prgHdrIdx)->p_offset,
						(exe->prgHdr+prgHdrIdx)->p_filesz);

					/* Give names to loadable sections
					 * (usually code and data). */
					if(exe->maps) {
						map=exe->maps;					
						while(map->next) map=map->next;
						map->next=calloc(1, sizeof(struct map_t));
						if(!map->next) {
							fprintf(stderr, "Error getting mem for exe map\n");
							errLoc=1;
						}
						map=map->next;
					}
					else {
						exe->maps=calloc(1, sizeof(struct map_t));
						if(!exe->maps) {
							fprintf(stderr, "Error getting mem for exe map\n");
							errLoc=1;
						}
						map=exe->maps;
					}
					if(!errLoc && map) {
						map->hdr=calloc(1, sizeof(Elf32_Phdr));
					}
					if(!errLoc && map && map->hdr) {
						memcpy(map->hdr, exe->prgHdr+prgHdrIdx, 
							sizeof(Elf32_Phdr));
						map->name=malloc(strlen(exe->path)+1);
						if(map->name) strcpy(map->name, exe->path);
					}
					break;
									
				default:
					break;
			}
		}
	}

	return errLoc;
}







//----------------------------------------------------------
// Start the binary executable and get some debuginfo
// out from the dynamic linker.
//----------------------------------------------------------
int exeNameDynamics(void) {
	const char *cmdPre="exec env LD_TRACE_LOADED_OBJECTS=1 ";
	const long ldOutBufLen=1024*16;
	char ldOutBuf[ldOutBufLen], *ldTmp;
	char file[PATH_MAX];	
	unsigned long addr;
	struct map_t *map;
	FILE *cmdStream;
	int errLoc, res;
	char *cmd;
	
	cmd=NULL;
	cmdStream=NULL;
	errLoc=0;
	memset(ldOutBuf, 0, ldOutBufLen);
	
	if(exe && exe->isDynamic) {
		// Create commandline
		cmd=malloc(strlen(cmdPre) + strlen(exe->path) + 2);
		if(cmd) {
			strcpy(cmd, cmdPre);
			strncat(cmd, exe->path, PATH_MAX-strlen(cmdPre)-1);
		}
		else {
			fprintf(stderr, "Error getting memory\n");
			errLoc=1;
		}
	
		// Start the dynamic linker
		if(!errLoc) {
			cmdStream=popen(cmd, "r");
			if(!cmdStream) {
				fprintf(stderr, "Error 1 checking out %s\n", exe->path);
				errLoc=1;
			}
		}

		/* Read stdout from dynamic linker */	
		if(!errLoc) {	
			ldTmp=ldOutBuf;
			while(ldTmp) {
				ldTmp=fgets(ldOutBuf, ldOutBufLen, cmdStream);			
				while(strchr(ldOutBuf, '\n')) *(strrchr(ldOutBuf, '\n'))='\0';	// Convert newline into null	

				if(strlen(ldOutBuf)) {
					res=sscanf(ldOutBuf, "%*s => %s (%lx)", file, &addr);
					if(res==0 || res==EOF) {
						fprintf(stderr, "Error parsing linker\n");
					}
					else if(addr>0 && strlen(file)) {
						
						/* Create program headers of our own. */
						if(exe->maps) {
							map=exe->maps;					
							while(map->next) map=map->next;
							map->next=calloc(1, sizeof(struct map_t));
							if(!map->next) {
								fprintf(stderr, 
									"Error getting mem for exe map\n");
								errLoc=1;
							}
							map=map->next;
						}
						else {
							exe->maps=calloc(1, sizeof(struct map_t));
							if(!exe->maps) {
								fprintf(stderr, 
									"Error getting mem for exe map\n");
								errLoc=1;
							}
							map=exe->maps;
						}
						if(!errLoc && map) {
							map->hdr=calloc(1, sizeof(Elf32_Phdr));
						}
						if(!errLoc && map && map->hdr) {
							map->hdr->p_vaddr=addr;
							map->hdr->p_flags = PF_R | PF_X;
							map->isDynamic=1;
							map->name=malloc(strlen(file)+1);
							if(map->name) strcpy(map->name, file);
						}
						file[0]='\0';
						addr=0;
					}
					ldOutBuf[0]='\0';											// Reset buffer			
				}
			}

			res=pclose(cmdStream);
			cmdStream=NULL;
			if(WIFEXITED(res)==0 || WIFSIGNALED(res) || WCOREDUMP(res)) {
				fprintf(stderr,"\nError 2 checking out %s\n", exe->path);
				errLoc=1;
			}
		}
	
		free(cmd);
	}
	
	return errLoc;
}




