/*
 *	coretrace - basic debugger for embedded Linux
 *	Copyright (C) 2004 Ronny L Nilsson
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
 *
 */

 
#include <sys/types.h>
#include <asm/string.h>
#include <signal.h>
#include <errno.h>
#include <sched.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/resource.h>									 	// getrlimit()

#ifdef DEBUG
#include <stdarg.h>
#endif

#include "main.h"
#include "core-linux.h"
#include "exe-elf.h"




//====== DEFINITIONS =================


//====== GLOBALS ==================
enum runmode_t runmode;											// Current state of program
char *corePath;
char *exePath;

static const char *progName="coretrace";
static const char *progVer="0.75beta1";


//====== PROTOTYPES ================
int parseArgs(int, char**);
int crasher1(void);
int crasher2(void);
int enableCoreFiles(void);



//----------------------------------------------------------
int main(int argc, char **argv) {
	int errLoc;
	
	errLoc=0;
	exe=NULL;
	core=NULL;
	runmode=RUNMODE_NONE;
	exePath=NULL;
	corePath=NULL;
	
	if(!errLoc) errLoc=parseArgs(argc, argv);	
	if(!errLoc) errLoc=coreInit();												// Tell kernel we want corefiles if we crash

	
	if(!errLoc) {
		switch(runmode) {
			case RUNMODE_NORMAL:
				if(!errLoc && exePath) errLoc=exeOpen(exePath);
				if(!errLoc) errLoc=exeGetSummary();								// Parse and get summary out of binary executable	
				if(!errLoc && corePath) errLoc=coreOpen(corePath);				// Open core from crashed app, change path if necessary!	
				if(!errLoc) {
					char *info=coreGetSummary();								// Get readable and useful info out 	
					if(info && strlen(info)) {									// 	of the corefile as text.	
						printf(info);
					}
					else {
						printf("No core summary available\n");
					}
				}																		
				coreClose();													// Always release resources and cleanup.	
				exeClose();
				break;
			
			case RUNMODE_CRASH:
				/* For testing, generate a ./core. */
				printf("Generating corefile....\n");
				errLoc=enableCoreFiles();
				if(errLoc) {
					printf("Aborted.\n");
				}
				else {
					fflush(stdout);												// Flush before crash	
					sched_yield();
					errLoc=crasher1();
				}
				break;

			case RUNMODE_VERSION:
				printf("%s v%s\n", progName, progVer);
				break;
			
			case RUNMODE_USAGE:
			case RUNMODE_NONE:
				printf("Usage: %s [options]\n", progName);
				printf("Core file analyser\n");
				printf("\n");
				printf("  -e [path]      Path to crashed binary executable\n");
				printf("  -c [path]      Path to core file\n");
				printf("  -v             Print version and exit\n");
				printf("  -h             This help\n");
				printf("  -x             Generate a crash, for testing purpose\n");
				break;
				
			default:
				break;
		}
	}

	return errLoc;
}



//----------------------------------------------------------
// Parse command line options
//----------------------------------------------------------
int parseArgs(int argc, char **argv) {
	int arg, errLoc;
	
	opterr=0;																	// Disable lib error msg's
	errLoc=0;
	
	while((arg=getopt(argc, argv, ":e:c:vhx")) !=-1 && !errLoc) {
		switch(arg) {
			case 'e':															// Executable path
				if(runmode==RUNMODE_NONE || runmode==RUNMODE_NORMAL) {
					runmode=RUNMODE_NORMAL;
					if(strlen(optarg)) exePath=optarg;
				}
				else {
					fprintf(stderr, "Option -e not allowed here\n");
					errLoc=1;
				}
				break;
			
			case 'c':															// Corefile path	
				if(runmode==RUNMODE_NONE || runmode==RUNMODE_NORMAL) {
					runmode=RUNMODE_NORMAL;
					if(strlen(optarg)) corePath=optarg;
				}
				else {
					fprintf(stderr, "Option -c not allowed here\n");
					errLoc=1;
				}
				break;
												
			case 'v':															// VERSION
				if(runmode==RUNMODE_NONE) {
					runmode=RUNMODE_VERSION;
				}
				else {
					fprintf(stderr, "Option -v not allowed here\n");
					errLoc=1;
				}
				break;
			
			case 'h':															// USAGE
				if(runmode==RUNMODE_NONE) {
					runmode=RUNMODE_USAGE;
				}
				else {
					fprintf(stderr, "Option -h not allowed here\n");
					errLoc=1;
				}
				break;
				
			case 'x':															// Crash and generate core	
				if(runmode==RUNMODE_NONE) {
					runmode=RUNMODE_CRASH;
				}
				else {
					fprintf(stderr, "Option -x not allowed here\n");
					errLoc=1;
				}
				break;
							
			case ':':															// Errorhandling
				fprintf(stderr, "Error, missing option ");						//  missing options.
				if(isprint(optopt)) fprintf(stderr, "to -%c", optopt);
				fprintf(stderr, "\n");
				fprintf(stderr, "Use -h for help\n");
				errLoc=1;
				break;
				
			default:															// Errorhandling
				fprintf(stderr, "Error, unknown option ");						//  unknown options.
				if(isprint(optopt)) fprintf(stderr, "-%c", optopt);
				fprintf(stderr, "\n");
				fprintf(stderr, "Use -h for help\n");
				errLoc=1;
				break;
		}
	
	}

	return errLoc;
}	




#ifdef DEBUG
//----------------------------------------------------------
// Debug printouts, enable them at compiletime
// with -DDEBUG in Makefile.
//----------------------------------------------------------
int coreDbg(const char *file, const char *function, long line, 
		const char *text, ...) {
	int len, res;
	va_list args;
	
	len=0;
	res=0;
	
	va_start(args, text);													
	res=vprintf(text, args);
	if(res>0) len+=res;
	va_end(args);
	
	return len;
}
#endif



//----------------------------------------------------------
// Tell kernel we want corefiles of size 
// <MAX_CORE_SIZE> (incase we crash). Returns 0 on 
// success and nonzero on errors.
//----------------------------------------------------------
int enableCoreFiles(void) {
	int errLoc, res;
	struct rlimit limitCore;

	errLoc=0;
	res=0;
	
	/* Get current limits. */
	res=getrlimit(RLIMIT_CORE, &limitCore);
	if(res==-1) {
		perror("Error getting core limits");
		errLoc=1;
	}

	/* Set new limits. If we're running as root
	 * set hard limit to maximum. */
	if(!errLoc) {
		if(geteuid()==0) limitCore.rlim_max=RLIM_INFINITY;
		if(limitCore.rlim_max >= MAX_CORE_SIZE) {
			limitCore.rlim_cur=MAX_CORE_SIZE;
		}
		else {
			fprintf(stderr, "Warning, can't set " 
				"core size as high as %u\n", MAX_CORE_SIZE);
			limitCore.rlim_cur=limitCore.rlim_max;
		}
	}
	if(!errLoc) {
		res=setrlimit(RLIMIT_CORE, &limitCore);
		if(res==-1) {
			perror("Error setting core limits");
			errLoc=1;
		}
	}

	return errLoc;
}



//----------------------------------------------------------
// Genereate core for testing purpose.
//----------------------------------------------------------
int crasher2(void) {
	/* Crash right here, use rand() to
	 * prevent compiler optimizations. */
	*(volatile int*)0=rand();		
	
	//raise(SIGQUIT);											// Crash by glibc	
	//printf("%s", (char*)1);									// Crash by glibc	
	
	return rand()+1;
}


//----------------------------------------------------------
// Genereate core for testing purpose.
//----------------------------------------------------------
int crasher1(void) {
	/* Dummy func to get another stackframe.
	 * We wan't something nice to trace. */
	if(crasher2()) return rand();
	else return crasher2();
}


